library flight_script;

import 'package:flight_script/vm/dictionary.dart';
import 'package:flight_script/vm/function.dart';
import 'package:flight_script/vm/host_function.dart';
import 'package:flight_script/vm/macro.dart';
import 'package:flight_script/vm/stack.dart';
import 'package:flight_script/vm/stdlib.dart';
import 'package:flight_script/vm/virtual_machine.dart';

import 'rpn/parser.dart';

export 'libs/dart_api.dart' show DartCompoundApi, DartSimpleApi;
export 'libs/lib.dart' show Library, LoadLibrary;
export 'vm/virtual_machine.dart' show VirtualMachine;

/// A vm for debugging efforts. Prints each evaluated value and the stack after
/// it's evaluated
/// We expect there to only be one stack listener, so we can extract the value
/// with a slightly hacky call
class DebugVm extends VirtualMachine {
  late final StackListener _stack;

  @override
  void pushListener(ValueListener listener) {
    if (listener is StackListener) {
      _stack = listener;
    }
    super.pushListener(listener);
  }

  @override
  dynamic eval(value) {
    final retval = super.eval(value);
    print(
        "${_valueString(value)}\t\t(${List.of(_stack.stack).map((it) => _valueString(it)).join(" ")})");
    return retval;
  }

  String _valueString(value) {
    if (value is DartFunction) {
      return "fn(~native)";
    }
    return value.toString();
  }
}

class Interpreter {
  final VirtualMachine vm;

  Interpreter({VirtualMachine? vm}) : this.vm = vm ?? VirtualMachine() {
    this.vm.pushListener(StackListener());
    this.vm.pushListener(DictionaryListener());
    this.vm.pushListener(HostFunctionListener());
    this.vm.pushListener(MacroListener());
    this.vm.pushListener(FlightFunctionListener());
    this.vm.loadStdLib();
  }

  List<dynamic> eval(String input) {
    final program = Parser().parse(input).toList();
    vm.eval(Macro(program));
    return vm.eval(Stack());
  }

  void bind(String name, DartFunction fn) => vm[name] = fn;
}
